/*
 * Copyright (c) 2022, NVIDIA CORPORATION.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <cudf/utilities/error.hpp>

#include <cstdint>
#include <vector>

namespace cudf {
namespace io {
namespace text {

/**
 * @brief stores offset and size used to indicate a byte range
 */
class byte_range_info {
 private:
  int64_t _offset;
  int64_t _size;

 public:
  constexpr byte_range_info() noexcept : _offset(0), _size(0) {}
  constexpr byte_range_info(int64_t offset, int64_t size) : _offset(offset), _size(size)
  {
    CUDF_EXPECTS(offset >= 0, "offset must be non-negative");
    CUDF_EXPECTS(size >= 0, "size must be non-negative");
  }

  constexpr byte_range_info(byte_range_info const& other) noexcept = default;
  constexpr byte_range_info& operator=(byte_range_info const& other) noexcept = default;

  [[nodiscard]] constexpr int64_t offset() { return _offset; }
  [[nodiscard]] constexpr int64_t size() { return _size; }
};

/**
 * @brief Create a collection of consecutive ranges between [0, total_bytes).
 *
 * Each range wil be the same size except if `total_bytes` is not evenly divisible by
 * `range_count`, in which case the last range size will be the remainder.
 *
 * @param total_bytes total number of bytes in all ranges
 * @param range_count total number of ranges in which to divide bytes
 * @return Vector of range objects
 */
std::vector<byte_range_info> create_byte_range_infos_consecutive(int64_t total_bytes,
                                                                 int64_t range_count);

/**
 * @brief Create a byte_range_info which represents as much of a file as possible. Specifically,
 * `[0, numeric_limit<int64_t>::max())`.
 *
 * @return `[0, numeric_limit<int64_t>::max())`
 */
byte_range_info create_byte_range_info_max();

}  // namespace text
}  // namespace io
}  // namespace cudf
