/*
 * Copyright (c) 2020-2022, NVIDIA CORPORATION.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <cstdio>
#include <cstdlib>
#include <filesystem>
#include <string>

#include <ftw.h>

#include <cudf/utilities/error.hpp>

/**
 * @brief RAII class for creating a temporary directory.
 *
 */
class temp_directory {
  std::string _path;

 public:
  temp_directory(const std::string& base_name)
  {
    std::string dir_template{std::filesystem::temp_directory_path().string()};
    if (auto env_p = std::getenv("WORKSPACE")) dir_template = env_p;

    dir_template += "/" + base_name + ".XXXXXX";
    auto const tmpdirptr = mkdtemp(const_cast<char*>(dir_template.data()));
    CUDF_EXPECTS(tmpdirptr != nullptr, "Temporary directory creation failure: " + dir_template);

    _path = dir_template + "/";
  }

  temp_directory& operator=(temp_directory const&) = delete;
  temp_directory(temp_directory const&)            = delete;
  temp_directory& operator=(temp_directory&&) = default;
  temp_directory(temp_directory&&)            = default;

  ~temp_directory() { std::filesystem::remove_all(std::filesystem::path{_path}); }

  /**
   * @brief Returns the path of the temporary directory
   *
   * @return string path of the temporary directory
   */
  [[nodiscard]] const std::string& path() const { return _path; }
};
